"""
http range detection based on
 https://github.com/menboku/musicsharer/blob/master/HTTPRangeServer.py
"""
import re
import socket

from SocketServer import ThreadingMixIn
from BaseHTTPServer import HTTPServer
from SimpleHTTPServer import SimpleHTTPRequestHandler

from os import fstat
from os.path import join

from guessit import guess_video_info

from touchandgo.settings import TMP_DIR


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
        pass


def serve_file(manager):

    class VideoHandler(SimpleHTTPRequestHandler):
        def handle_one_request(self, *args, **kwargs):
            try:
                return SimpleHTTPRequestHandler.handle_one_request(self, *args,
                                                                   **kwargs)
            except:
                pass

        def do_GET(self):
            f = self.send_head()
            if f:
                if self.range_from is not None and self.range_to is not None:
                    self.copy_chunk(f, self.wfile)
                else:
                    self.copyfile(f, self.wfile)
                f.close()

        def get_video_path(self):
            video = manager.video_file
            video_path = join(TMP_DIR, video[0])
            return video_path

        def copy_chunk(self, in_file, out_file):
            """
            Copy a chunk of in_file as dictated by self.range_[from|to]
            to out_file.
            NB: range values are inclusive so 0-99 => 100 bytes
            Neither of the file objects are closed when the
            function returns.  Assumes that in_file is open
            for reading, out_file is open for writing.
            If range_tuple specifies something bigger/outside
            than the size of in_file, out_file will contain as
            much content as matches.  e.g. with a 1000 byte input,
            (500, 2000) will create a 500 byte long file
            (2000, 3000) will create a zero length output file
            """

            in_file.seek(self.range_from)
            # Add 1 because the range is inclusive
            left_to_copy = 1 + self.range_to - self.range_from

            bytes_copied = 0
            while bytes_copied < left_to_copy:
                read_buf = in_file.read(left_to_copy)
                if len(read_buf) == 0:
                    break
                out_file.write(read_buf)
                bytes_copied += len(read_buf)
            return bytes_copied

        def send_head(self):
            """Common code for GET and HEAD commands.

            This sends the response code and MIME headers.

            Return value is either a file object (which has to be copied
            to the outputfile by the caller unless the command was HEAD,
            and must be closed by the caller under all circumstances), or
            None, in which case the caller has nothing further to do.


            """
            path = self.get_video_path()
            try:
                # Always read in binary mode. Opening files in text mode may
                # cause newline translations, making the actual size of the
                # content transmitted *less* than the content-length!
                f = open(path, 'rb')
            except IOError:
                self.send_error(404, "File not found")
                return None

            fs = fstat(f.fileno())
            total_length = fs[6]
            try:
                self.range_from, self.range_to = parse_range_header(
                    self.headers.getheader("Range"), total_length)
            except InvalidRangeHeader:
                # Just serve them the whole file, although it's possibly
                # more correct to return a 4xx error?
                #logging.warning("Range header parsing failed, "
                #                "serving complete file")
                self.range_from = self.range_to = None

            self.send_header("Accept-Ranges", "bytes")
            if self.range_from is not None or self.range_to is not None:
                self.send_response(206)
            else:
                self.send_response(200)
            guess = guess_video_info(path, info=['filename'])
            self.send_header("Content-Type", guess['mimetype'])
            if self.range_from is not None or self.range_to is not None:
                # TODO: Should also check that range is within the file size
                self.send_header("Content-Range",
                                 "bytes %d-%d/%d" % (self.range_from,
                                                     self.range_to,
                                                     total_length))
                # Add 1 because ranges are inclusive
                self.send_header("Content-Length",
                                (1 + self.range_to - self.range_from))
            else:
                self.send_header("Content-Length", str(total_length))
            self.send_header("Last-Modified",
                             self.date_time_string(fs.st_mtime))
            self.end_headers()
            return f

        def finish(self, *args, **kwargs):
            try:
                if not self.wfile.closed:
                    self.wfile.flush()
                    self.wfile.close()
            except Exception:
                pass
            self.rfile.close()

    def run(server_class=ThreadedHTTPServer,
            handler_class=SimpleHTTPRequestHandler):
        print("serving on http://localhost:%s" % manager.port)
        server_address = ('0.0.0.0', manager.port)
        httpd = server_class(server_address, handler_class)
        httpd.serve_forever()

    run(handler_class=VideoHandler)


class InvalidRangeHeader(Exception):
    pass


def parse_range_header(range_header, total_length):
    """
    Return a 2-element tuple containing the requested range offsets
    in bytes.
    - range_header is the HTTP header sans the "Range:" prefix
    - total_length is the length in bytes of the requested resource
      (needed to calculate offsets for a 'n bytes from the end' request
    If no Range explicitly requested, returns (None, None)
    If Range header could not be parsed, raises InvalidRangeHeader
    (which could either be handled as a user
    request failure, or the same as if (None, None) was returned
    """
    # range_header = self.headers.getheader("Range")
    if range_header is None or range_header == "":
        return (None, None)
    if not range_header.startswith("bytes="):
        # logging.error("Don't know how to parse Range: %s [1]" %
        #                (range_header))
        raise InvalidRangeHeader("Don't know how to parse non-bytes Range: %s" %
                                 (range_header))
    regex = re.compile(r"^bytes=(\d*)\-(\d*)$")
    rangething = regex.search(range_header)
    if rangething:
        r1 = rangething.group(1)
        r2 = rangething.group(2)
        #logging.debug("Requested range is [%s]-[%s]" % (r1, r2))

        if r1 == "" and r2 == "":
            # logging.warning("Requested range is meaningless")
            raise InvalidRangeHeader("Requested range is meaningless")

        if r1 == "":
            # x bytes from the end of the file
            try:
                final_bytes = int(r2)
            except ValueError:
                raise InvalidRangeHeader("Invalid trailing range")
            return (total_length-final_bytes, total_length - 1)

        try:
            from_val = int(r1)
        except ValueError:
            raise InvalidRangeHeader("Invalid starting range value")
        if r2 != "":
            try:
                end_val = int(r2)
            except ValueError:
                raise InvalidRangeHeader("Invalid ending range value")
            return (from_val, end_val)
        else:
            return (from_val, total_length - 1)
    else:
        raise InvalidRangeHeader("Don't know how to parse Range: %s" %
                                 (range_header))





#! /usr/bin/env python
import argparse
import signal

from os import kill
from os.path import abspath, dirname, join

from subprocess import PIPE, STDOUT, Popen
from time import sleep
from flask import Flask, redirect, render_template, Response
from jinja2 import FileSystemLoader

from helpers import get_interface, get_lock_diff, LOCKFILE, set_config_dir
from history import History
from lock import Lock


def serve(py_exec=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("--python", default="python")
    args = parser.parse_args()

    py_exec = args.python

    app = Flask("touchandgo")
    template_path = join(dirname(__file__), "templates")
    app.jinja_loader = FileSystemLoader(template_path)

    @app.route("/<name>")
    @app.route("/<name>/<season>/<episode>")
    def redirect_to(name, season=None, episode=None):
        interface = get_interface()
        path = abspath(__file__)
        dir_path = dirname(path)
        exec_ = join(dir_path, "touchandgo.py")
        command = '%s %s \"%s\" ' % (py_exec, exec_, name)
        if season is not None:
            command += "%s %s " % (season, episode)
        command += "--daemon"
        process = Popen(command, shell=True, stdout=PIPE, stderr=STDOUT)
        sleep(1)
        while get_lock_diff() < 10:
            sleep(1)
        port = 8888
        return redirect("http://%s:%s" % (interface, port))

    @app.route("/l/<name>/<season>/<episode>")
    @app.route("/l/<name>/<season>/<episode>/")
    def list_(name, season, episode):
        url = "http://%s:5000/%s/%s" % (get_interface(), name, season)
        episode = int(episode)

        template = app.jinja_env.get_template("m3u.j2")
        ret = template.render(episodes=range(episode, episode + 10),
                              series=name, season=episode, url=url)
        return Response(response=ret, status=200,
                        mimetype="application/x-mpegurl",
                        content_type="application/x-mpegurl")

    @app.route("/kill")
    def kill_():
        lock = Lock(LOCKFILE)
        kill(lock.get_pid(), signal.SIGQUIT)
        return "OK"

    @app.route("/")
    def main():
        series = []
        keys = []
        items = History.many(sorted="-date,-season,-episode")
        for item in items:
            key = (item.name, item.season is not None)
            if key not in keys:
                series.append(item)
                keys.append(key)

        return render_template("main.html", items=series)


    set_config_dir()
    app.debug = True
    app.run(host="0.0.0.0")


if __name__ == '__main__':
    serve()



import thread

from os import system, _exit
from os.path import join
from time import sleep
from datetime import datetime

from babelfish import Language
from guessit import guess_video_info
from libtorrent import add_magnet_uri, session
from subliminal import download_best_subtitles
from subliminal.subtitle import get_subtitle_path
from subliminal.video import Video

from touchandgo.constants import STATES
from touchandgo.helpers import is_port_free, get_free_port
from touchandgo.settings import DEBUG, TMP_DIR
from touchandgo.stream_server import serve_file


class DownloadManager(object):
    def __init__(self, magnet, port=None, sub_lang=None, serve=False):
        self.magnet = magnet
        if port is None:
            port = 8888
        port = int(port)
        if not is_port_free(port):
            port = get_free_port()

        self.port = port
        self.sub_lang = sub_lang
        self.serve = serve

        # number of pieces to wait until start streaming
        self.piece_st = 4
        # we are waiting untill all the first peices are downloaded
        self.holding_stream = True
        # the biggest file which is supposed to be a video file
        self._video_file = None
        self.callback = serve_file

        self.init_handle()

    def init_handle(self):
        params = {
            "save_path": TMP_DIR,
            #"allocation": "compact"
            }
        self.session = session()
        self.handle = add_magnet_uri(self.session, str(self.magnet), params)

    def start(self):
        self.start_time = datetime.now()
        print("Downloading metadata")
        while not self.handle.has_metadata():
            sleep(.1)
        print("Starting download")
        self.session.listen_on(6881, 6891)
        self.session.start_dht()

        chunks_strat = self.initial_strategy()

        try:
            while True:
                if not self.handle.is_seed():
                    self.strategy_master(chunks_strat)
                elif self.holding_stream:
                    self.holding_stream = False
                    self.stream_video()
                print("\n" * 80)
                if DEBUG:
                    self.defrag()
                self.stats()
                sleep(1)
        except KeyboardInterrupt:
            _exit(0)

    @property
    def video_file(self):
        if self._video_file is None:
            self._video_file = self.get_biggest_file()

        return self._video_file

    def get_biggest_file(self):
        info = self.handle.get_torrent_info()
        biggest_file = ("", 0)
        files = info.files()
        for file_ in files:
            if file_.size > biggest_file[1]:
                biggest_file = [file_.path, file_.size]

        return biggest_file

    def run_vlc(self):
        command = "vlc http://localhost:%s -q" % self.port
        if self.sub_lang is not None:
            subtitle = self.get_subtitle()
            if subtitle is not None:
                command += " --sub-file %s" % subtitle
        try:
            system(command)
        except KeyboardInterrupt:
            pass
        _exit(0)

    def strategy_master(self, chunks_strat):
        status = self.handle.status()
        pieces = status.pieces
        first_n = pieces[:self.piece_st]

        if all(first_n):
            self.handle.set_sequential_download(False)
            pieces_strat = pieces[self.piece_st:self.piece_st + chunks_strat]
            if self.holding_stream or all(pieces_strat):
                if not self.holding_stream:
                    self.piece_st += chunks_strat
                else:
                    if self.callback is not None:
                        self.stream_video()

                for i in range(self.piece_st, self.piece_st + chunks_strat):
                    self.handle.piece_priority(i, 7)
                    self.handle.set_piece_deadline(i, 10000)

                for i in range(self.piece_st+chunks_strat,
                               self.piece_st + chunks_strat*2):
                    self.handle.piece_priority(i, 5)
                    self.handle.set_piece_deadline(i, 20000)
                self.holding_stream = False

    def stream_video(self):
        thread.start_new_thread(self.callback, (self, ))
        if not self.serve:
            thread.start_new_thread(self.run_vlc, ())

    def defrag(self):
        status = self.handle.status()
        numerales = ""
        for i, piece in enumerate(status.pieces):
            numeral = "#" if piece else " "
            numeral += str(self.handle.piece_priority(i))
            numerales += numeral
        print(numerales)

    def stats(self):
        status = self.handle.status()
        print '%.2f%% complete (down: %.1f kb/s up: %.1f kB/s peers: %d) %s' % \
            (status.progress * 100, status.download_rate / 1000,
             status.upload_rate / 1000, status.num_peers, STATES[status.state])
        print "Elapsed Time", datetime.now() - self.start_time

    def initial_strategy(self):
        self.handle.set_sequential_download(True)
        status = self.handle.status()

        for i in range(self.piece_st):
            self.handle.piece_priority(i, 7)
            self.handle.set_piece_deadline(i, 10000)

        """
        last_piece = len(status.pieces) - 1
        for i in range(last_piece-1, last_piece+1):
            self.handle.piece_priority(i, 7)
            self.handle.set_piece_deadline(i, 10000)
        """

        return len(status.pieces) / 25

    def get_subtitle(self):
        print("Downloading subtitle")
        video_file = self.video_file
        filepath = join(TMP_DIR, video_file[0])
        guess = guess_video_info(filepath, info=['filename'])
        video = Video.fromguess(filepath, guess)
        video.size = video_file[1]
        subtitle = download_best_subtitles([video], {Language(self.sub_lang)},
                                           single=True)
        if not len(subtitle):
            subtitle = None
        else:
            subtitle = get_subtitle_path(join(TMP_DIR, video.name))
        return subtitle




import os
import signal
import socket

from daemon import DaemonContext
from datetime import datetime
from lock import Lock
from os import mkdir
from os.path import getmtime, exists

from netifaces import interfaces, ifaddresses

from ojota import set_data_source


LOCKFILE = "/tmp/touchandgo"


def get_free_port():
    socket_ = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    socket_.bind(('localhost', 0))
    addr, port = socket_.getsockname()
    socket_.close()
    return port

def is_port_free(port):
    free = True
    socket_ = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        socket_.bind(('localhost', port))
    except socket.error:
        free = False
    socket_.close()

    return free



def get_interface():
    for ifaceName in interfaces():
        addresses = ifaddresses(ifaceName)
        for address in addresses.values():
            for item in address:
                if item.get('netmask') is not None and \
                        not item['addr'].startswith("127") and \
                        not item['addr'].startswith(":") and \
                        len(item['addr']) < 17:
                    return item['addr']


def is_process_running(process_id):
    try:
        os.kill(process_id, 0)
        return True
    except OSError:
        return False


def daemonize(args, callback):
    with DaemonContext():
        create_process = False
        lock = Lock(LOCKFILE, os.getpid(), args.name, args.sea_ep[0],
                    args.sea_ep[1], args.port)
        if lock.is_locked():
            lock_pid = lock.get_pid()
            if not lock.is_same_file(args.name, args.sea_ep[0],
                                     args.sea_ep[1]) \
                    or not is_process_running(lock_pid):
                try:
                    os.kill(lock_pid, signal.SIGQUIT)
                except OSError:
                    pass
                except TypeError:
                    pass
                lock.break_lock()
                create_process = True
        else:
            create_process = True

        if create_process:
            lock.acquire()
            callback(args.name, season=args.sea_ep[0], episode=args.sea_ep[1],
                     serve=True, port=args.port)
            lock.release()


def get_lock_diff():
    timediff = 0
    try:
        now = datetime.now()
        timediff = now - datetime.fromtimestamp(getmtime(LOCKFILE + ".lock"))
        timediff = timediff.total_seconds()
    except OSError:
        pass
    return timediff


def set_config_dir():
    data_folder = "%s/.touchandgo" % os.getenv("HOME")
    if not exists(data_folder):
        mkdir(data_folder)

    set_data_source(data_folder)




#! /usr/bin/env python
import argparse

from time import time
from torrentmediasearcher import TorrentMediaSearcher

from touchandgo.helpers import daemonize, set_config_dir
from touchandgo.history import History
from touchandgo.download import DownloadManager


def watch(name, season=None, episode=None, sub_lang=None, serve=False,
          quality=None, port=None):

    def get_magnet(results):
        print("Processing magnet link")
        magnet = results['magnet']
        print(magnet)
        manager = DownloadManager(magnet, port=port, serve=serve,
                                  sub_lang=sub_lang)
        manager.start()
        set_config_dir()

        history = History(date=int(time()), name=name, season=season,
                          episode=episode)
        history.save()
        history.update()

    print("Searching torrent")
    search = TorrentMediaSearcher
    if season is None and episode is None:
        search.request_movie_magnet('torrentproject', name,
                                    callback=get_magnet, quality=quality)
    else:
        if quality is None:
            quality = 'normal'
        search.request_tv_magnet(provider='eztv', show=name,
                                 season=int(season), episode=int(episode),
                                 quality=quality, callback=get_magnet)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("name")

    parser.add_argument("sea_ep", nargs='*', default=[None, None])
    parser.add_argument("--sub", nargs='?', default=None)
    parser.add_argument("--serve", action="store_true")
    parser.add_argument("--quality", nargs='?', default=None)
    parser.add_argument("--daemon", action="store_true")
    parser.add_argument("--port", "-p", default="8888")
    parser.add_argument("--season", action="store_true")
    args = parser.parse_args()

    if args.daemon:
        daemonize(args, watch)
    else:
        episode = int(args.sea_ep[1]) if args.sea_ep[1] is not None else None
        play_next_episode = True
        while play_next_episode:
            watch(args.name, season=args.sea_ep[0], episode=episode,
                  sub_lang=args.sub, serve=args.serve, quality=args.quality,
                  port=args.port)
            episode += 1
            play_next_episode = args.season

if __name__ == '__main__':
    main()
